#pylint: disable=line-too-long,import-error,too-few-public-methods
"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import json

import boto3



class GAV4Writer:
    """GAV4Writer Class"""

    def __init__(self, bucket, folder_path, profile_name=None):
        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)
        self.s3_resource = self.boto3_session.resource("s3")
        self.bucket = bucket
        self.folder_path = folder_path
        self.data = None

    def write_to_s3(self, payload):
        """
        This pulls the yielded dataset from the GA reader in a manner
        that consumes the dataset of the given view_id and date,
        and writes it to s3 so that duplication does not occur.
        :param payload: This is a key value object that looks like:
                        {
                            "data": list(),
                            "date": string,
                            "property_id": string
                        }
        """

        # confirm the payload keys are matching accurately with what is expected
        if list(payload.keys()).sort() != ["data", "date", "property_id"].sort():
            raise KeyError("Invalid payload")

        _property_id = payload["property_id"]
        _date = payload["date"].replace("-", "")
        _data = payload["data"]
        write_path = f"{self.folder_path}/{_property_id}/{_date}.json"
        if _data:
            print(
                f"Writing data to s3://{self.bucket}/{write_path} partitioned by property_id and date."
                )
            self.s3_resource.Object(self.bucket, write_path).put(Body=json.dumps(_data))
            self.data = {}



"""MODULE TO HELP WITH PROCESSING FILES"""
# pylint: disable=unused-argument
import json
import yaml


def load_file(file_location: str, fmt: str = None) -> dict:
    """
    Gathers file data from json or yaml.
    """
    config: dict = {}
    if file_location.strip().rsplit(".", maxsplit=1)[-1] not in ["json", "yml", "yaml"]:
        raise TypeError("Wrong file type provided! Expecting only json and yaml files")

    file_location = str(file_location).strip()
    if file_location.endswith("yml"):
        with open(file_location, mode="r", encoding="utf8") as yaml_file:
            config = yaml.safe_load(yaml_file)
    if file_location.endswith("json"):
        with open(file_location, mode="r", encoding="utf8") as json_file:
            config = json.load(json_file)

    return config



# pylint: disable=ungrouped-imports
import os
import warnings
from datetime import datetime
from pyspark.sql import functions as F
import pyspark.sql.types as T


def user_defined_timestamp(date_col, date_format: str = "%Y%m%d"):
    """
    (UDF) convert datetime column using python datetime

    args:
        date_col (pyspark.sql.functions.col): datetime column to format
        date_format (str): a datetime valid string format
    return
        (pyspark.sql.functions.col)
    """
    _date = datetime.strptime(date_col, "%Y-%m-%d")
    return _date.strftime(date_format)


def advanced_writer(*, dataframe, output_data_config: dict = None, **kwargs):
    """
    Constructs a write function configuration and returns this.

    note:
        - on return write_function.csv() or write_function.json() or write_function.parquet()
    """
    # pylint: disable=too-many-locals, too-many-branches, too-many-statements

    if not isinstance(output_data_config, dict):
        raise TypeError("output_data_config was not of type dict.")

    output_path = output_data_config.get("output_path", None)
    if output_path is None:
        raise TypeError("output_path is required but was not found")

    output_type = output_data_config.get("output_type", None)
    if output_type is None:
        raise TypeError("output_type is required but was not found")

    partition_config = kwargs.get("partition_config", {})
    if not isinstance(partition_config, dict):
        raise TypeError("partition_config was not of type dict.")

    partitions = partition_config.get("partitions", None)
    partition_by = partition_config.get("partition_by", None)
    partition_type = partition_config.get("partition_type", None)

    bucket_config = kwargs.get("bucket_config", {})
    if not isinstance(bucket_config, dict):
        raise TypeError("bucket_config was not of type dict.")

    bucket_type = bucket_config.get("bucket_type", None)
    buckets = bucket_config.get("buckets", None)

    compression_type = kwargs.get("compression_type", None)
    overwrite = kwargs.get("overwrite", False)
    header = kwargs.get("header", True)

    # output from this step is a dataframe writer
    if partition_type in ["key", "random"]:
        if partition_by is None:
            raise Exception("To perform repartitioning 'partition_by' is required.")

        col_structure = []
        for column in partition_by:
            col_structure.append(F.col(column))

        if partition_type == "random":
            col_structure.append(F.rand)

        dataframe = dataframe.repartition(partitions, *col_structure)
        if bucket_type == "size":
            # this will be ignored in the case where partitions are added
            dataframe_writer = dataframe.write.option(
                "maxRecordsPerFile", buckets
            ).partitionBy(*partition_by)
        else:
            dataframe_writer = dataframe.write.partitionBy(*partition_by)

    elif bucket_type == "size":
        # this will be ignored in the case where partitions are added
        dataframe_writer = dataframe.write.option("maxRecordsPerFile", buckets)
    else:
        warnings.warn(
            (
                "Defaulting to standard writer. "
                "Either set a num_partitions or max_size "
                "to bucket for additional options."
            )
        )
        dataframe_writer = dataframe.write

    if overwrite is True:
        dataframe_writer = dataframe_writer.mode("overwrite")

    dataframe_writer = dataframe_writer.option("header", header)

    if compression_type is not None:
        dataframe_writer = dataframe_writer.option("compression", compression_type)

    if output_type == "csv":
        dataframe_writer.csv(output_path)
    elif output_type == "json":
        dataframe_writer.json(output_path)
    elif output_type == "parquet":
        dataframe_writer.parquet(output_path)
    else:
        raise Exception(
            f"Output type = {output_type} not supported. "
            "Only support csv and json at current."
        )


def advanced_writer_w_custom_partition_naming(
    *, dataframe, output_data_config: dict, **kwargs
):
    """
    SDC custom pyspark writer that extends advanced_writer
    to offer special date partitioning which includes a more
    well formatted set of directory names for single level partition.
    """
    # pylint: disable=too-many-locals, too-many-function-args

    if not isinstance(output_data_config, dict):
        raise TypeError("output_data_config was not of type dict.")

    partition_config = kwargs.get("partition_config", {})
    if not isinstance(partition_config, dict):
        raise TypeError("partition_config was not of type dict.")

    partition_by = partition_config.get("partition_by", None)
    partition_type = partition_config.get("partition_type", None)
    date_partition_format = partition_config.get("date_partition_format", None)

    bucket_config = kwargs.get("bucket_config", {})
    if not isinstance(bucket_config, dict):
        raise TypeError("bucket_config was not of type dict.")

    if partition_type == "date":
        output_path = output_data_config.get("output_path", None)
        if output_path is None:
            raise KeyError("output_path is required but was not found")

        output_type = output_data_config.get("output_type", None)
        if output_type is None:
            raise KeyError("output_type is required but was not found")

        if not isinstance(partition_by, str):
            raise TypeError(
                "For partition_type == date, "
                "the 'partition_by' field should be name of date field."
            )

        if date_partition_format is None:
            raise KeyError(
                "For partition_type == date, "
                "'date_partition_format' field is required but missing"
            )

        # pop partition_config from kwargs
        kwargs.pop("partition_config")

        # create udf to format date
        user_defined_timestamp_udf = F.udf(user_defined_timestamp, T.StringType())

        # create tmp_partition_field
        dataframe = dataframe.withColumn(
            "tmp_partition_field",
            user_defined_timestamp_udf(partition_by, F.lit(date_partition_format)),
        )

        partition_groups = list(
            dataframe.select("tmp_partition_field")
            .distinct()
            .rdd.map(lambda r: r[0])
            .collect()
        )

        for partition in partition_groups:
            partioned_dataframe = dataframe.filter(
                F.col("tmp_partition_field") == partition
            ).drop("tmp_partition_field")

            advanced_writer(
                dataframe=partioned_dataframe,
                output_data_config={
                    "output_path": os.path.join(output_path, partition),
                    "output_type": output_type,
                },
                **kwargs,
            )
    else:
        advanced_writer(dataframe, output_data_config=output_data_config, **kwargs)



"""
    CUSTOM FALCON READER CLASSES
"""
# pylint: disable=too-few-public-methods,import-error,unused-import
import os
from typing import Generator, List
from datetime import datetime, timedelta

import requests

from sdc_dp_helpers.api_utilities.date_managers import date_handler, date_range
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler


class CustomFalconReader:
    """
    Custom Falcon Reader
    """

    def __init__(self, creds_file: str, config_file=None):
        """creds_file has crecdentials config_file has api pull configs"""
        self._creds: dict = load_file(creds_file, "yml")
        self._config: dict = load_file(config_file, "yml")
        self.request_session: requests.Session = requests.Session()
        self.base_url: str = "https://api.falcon.io/"
        self.steps = 50
        self.start = 0
        self.curr_channel_ids = []

    @retry_handler(exceptions=ConnectionError, total_tries=5)
    def _get_channel_ids(self) -> List[str]:
        """
        Gather all available channel ids.
        """
        print("GET: channel ids.")
        endpoint_url = f"channels?apikey={self._creds['api_key']}"
        url = f"https://api.falcon.io/{endpoint_url}"
        response = self.request_session.get(url=url, params={"limit": "9999"})

        response_data = response.json()
        if response.status_code == 200:
            channel_ids = set()
            for item in response_data.get("items", []):
                channel_ids.add(item["id"])

            return list(channel_ids)

        raise ConnectionError(
            f"Falcon API failed to return channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=900)
    def _content_metrics_by_channel_id(self, date: str, channel_ids: list) -> list:
        """Gets the content metrics by channel id
        SEE: https://falconio.docs.apiary.io/ \
        #reference/channel-api/get-facebook-and-instagram-page-insights-metrics\
        # /get-content-metrics-by-channel-ids
        :param: channel_id - integer
        :param: date - string pushed as '2022-07-20'
        :returns: list of dictionaries
        """
        dataset: list = []
        endpoint_url: str = (
            f"measure/api/v1/content/metrics?apikey={self._creds['api_key']}"
        )
        offset: int = 0
        limit = self._config.get("limit", 9999)
        while True:
            print(
                f"INFO: channel id index: {self.start}, channel ids: "
                f"{len(channel_ids)}, date: {date}, offset: {offset}."
            )
            response = self.request_session.post(
                url=f"{self.base_url}{endpoint_url}",
                headers={"Content-Type": "application/json"},
                json={
                    "metrics": self._config.get("metrics", []),
                    "channels": channel_ids,
                    "since": date,
                    "until": date,
                    "postsSince": date,
                    "postsUntil": date,
                    "direction": self._config.get("direction", "ASC"),
                    "limit": limit,
                    "offset": offset,
                },
            )
            status_code: int = response.status_code
            reason: str = response.reason
            if status_code == 200:
                results: list = response.json()
                result_count: int = len(results)

                if result_count == 0:
                    break  # we will exit the loop because no data is got back

                if result_count < limit:
                    offset += result_count
                else:
                    offset += limit

                dataset.extend(results)
            elif status_code == 414 and "Request-URI Too Long" in reason:
                self.steps = self.steps - 20
                self.curr_channel_ids = self.curr_channel_ids[self.start : self.steps]
                print(f"len of channel_ids {len(channel_ids)}")
                raise TimeoutError(
                    "Request-URI Too Long. Reducing number of channel_ids by 20"
                )
            elif reason == "Not Found" and status_code == 404:
                print(f"No data for channel ids: {channel_ids}, skipping.\n")
                break
            elif status_code == 429:
                raise TimeoutError("Rolling Window Quota [429] reached.")
            else:
                raise ConnectionError(f"Reason: {reason}, Status: {status_code}.")

        return dataset

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=900)
    def _published_posts_by_channel_id(self, date: str, channel_ids: list) -> list:
        """Gets the published posts by channel id
        SEE: https://falconio.docs.apiary.io/
        #reference/content-api/get-content-metrics-by-channel-ids
        :param: channel_id - integer
        :param: date - string '2022-07-20' but converted to isoformat '2022-07-20T00:00:00'
        :returns: list of dictionaries
        """

        dataset: list = []
        endpoint_url: str = f"publish/items?apikey={self._creds['api_key']}"
        limit = self._config.get("limit", 2000)
        start_date = datetime.strptime(date, "%Y-%m-%d").isoformat()
        end_date = (datetime.strptime(date, "%Y-%m-%d") + timedelta(days=1)).isoformat()
        while endpoint_url:
            print(
                f"INFO: channel id index: {self.start}, "
                f"channel id: {channel_ids},  date: {date}, offset: {len(dataset)}."
            )
            response = self.request_session.get(
                url=f"{self.base_url}{endpoint_url}",
                headers={"Content-Type": "application/json"},
                params={
                    "channels": channel_ids,
                    "since": start_date,
                    "until": end_date,
                    "networks": self._config["networks"],
                    "statuses": self._config.get("statuses", "published"),
                    "limit": limit,
                },
            )
            status_code: int = response.status_code
            reason: str = response.reason
            if response.status_code == 200:
                results: dict = response.json()
                items_data: list = results.get("items", [])
                dataset.extend(items_data)

                endpoint_url = results.get("next", {"href": None}).get("href")

                if len(items_data) == 0 or endpoint_url is None:
                    break

            elif reason == "Not Found" and status_code == 404:
                print(f"No data for channel id: {channel_ids}, skipping.\n")
                break
            elif status_code == 429:
                raise TimeoutError("Rolling Window Quota [429] reached.")
            else:
                raise ConnectionError(f"Reason: {reason}, Status: {status_code}.")

        return dataset

    def _query_handler(self, date: str, channel_ids: list) -> list:
        """Interface to decide the api call to make"""
        endpoint_name = self._config.get("endpoint_name", None)
        self.start = 0
        dataset = []
        if not endpoint_name:
            raise Exception(
                "Please specify endpoint in the configs i.e content_metrics or published_posts"
            )
        if endpoint_name == "content_metrics":
            channel_len = len(channel_ids)
            for _ in range(0, channel_len, self.steps):
                self.curr_channel_ids = channel_ids[
                    self.start : self.start + self.steps
                ]
                results = self._content_metrics_by_channel_id(
                    date, self.curr_channel_ids
                )
                dataset.extend(results)
                self.start += self.steps
        elif endpoint_name == "published_posts":
            for channel_id in channel_ids:
                self.curr_channel_ids = [channel_id]
                results = self._published_posts_by_channel_id(date, [channel_id])
                dataset.extend(results)
                self.start += 1
        else:
            raise Exception(f"Invalid endpoint_name {endpoint_name}")

        return dataset

    def run_query(self) -> Generator[dict, None, None]:
        """
        Get metrics by channel Id context returns a request session with the ability
        to page with offsets.
        Content (or Post level) contains all metrics about your specific piece of
        content (posts). Here you will find impressions, reach, likes,
        shares and other metrics that show how well your specific post has performed.
        https://falconio.docs.apiary.io/reference/content-api/get-copied-content.
        """
        channel_ids: list = self._get_channel_ids()
        # channel_ids = ["114215121990330", "58283981268"]
        print(f"Attempting to gather metrics from {len(channel_ids)} channel ids.")
        for date in date_range(
            start_date=date_handler(self._config.get("since", None), "%Y-%m-%d"),
            end_date=date_handler(self._config.get("until", None), "%Y-%m-%d"),
        ):
            payload: list = []
            payload = self._query_handler(date=date, channel_ids=channel_ids)
            if len(payload) > 0:
                yield {
                    "networks": self._config["networks"],
                    "date": date.replace("-", ""),
                    "data": payload,
                }
            else:
                print(
                    f"No data for endpoint {self._config['endpoint_name']} for date : {date}"
                )
